/* integer.c   by Michael Thorpe   2017-04-12 */

/* If this is defined, we forbid multiplication operands that won't fit in a
 * signed int and require that a signed int be no more than half the size of
 * a signed long. Otherwise, we require that long long be at least twice the
 * size of long. If neither of those work for you, you'll need to implement
 * your own multiplication overflow detecting algorithm. */
#define RESTRICTED_MULTIPLICATION

#include <ctype.h>
#include <limits.h>
#include <stdio.h>
#include "lisp.h"

static obj *divideby0error;
static obj *overflowerror;

obj trueobj={&integer_objtype,{refcount:1},{i:1}};

obj *add(obj *a,obj *b) {
	signed long result;
	int overflow=0;
	obj *r;

	if(a->type != &integer_objtype) {
		DECREF(b);
		return(throwtypeerror(a));
	}
	if(b->type != &integer_objtype) {
		DECREF(a);
		return(throwtypeerror(b));
	}
/* FIXME: This has undefined behavior on overflow: */
	result=a->value.i+b->value.i;
	if((a->value.i^result)<0 && (b->value.i^result)<0)
		overflow=1;
	DECREF(a);
	DECREF(b);
	r=new_object(&integer_objtype);
	if(r) {
		r->value.i=result;
		if(overflow)
			return(throw(INCREF(overflowerror),r));
	}
	return(r);
}

obj *bitselect(obj *a,obj *if0,obj *if1) {
	obj *r;

	if(a->type != &integer_objtype) {
		DECREF(if0);
		DECREF(if1);
		return(throwtypeerror(a));
	}
	if(if0->type != &integer_objtype) {
		DECREF(a);
		DECREF(if1);
		return(throwtypeerror(if0));
	}
	if(if1->type != &integer_objtype) {
		DECREF(a);
		DECREF(if0);
		return(throwtypeerror(if1));
	}
/*
 * For each bit in a, selects the corresponding bit in if0 if it's 0 or
 * the corresponding bit in if1 if it's 1.  For instance:
 *   (bitwise-not a) is (bitselect a -1 0)
 *   (bitwise-and a b) is (bitselect a 0 b)
 *   (bitwise-or a b) is (bitselect a -1 b)
 *   (bitwise-xor a b) is (bitselect a b (bitwise-not b))
*/
	r=new_object(&integer_objtype);
	if(r)
		r->value.i=(~a->value.i & if0->value.i) | (a->value.i & if1->value.i);
	DECREF(a);
	DECREF(if0);
	DECREF(if1);
	return(r);
}

/* Arithmetic shift right */
obj *asr(obj *a,obj *b) {
	signed long result;
	obj *r;

	if(a->type != &integer_objtype) {
		DECREF(b);
		return(throwtypeerror(a));
	}
	if(b->type != &integer_objtype) {
		DECREF(a);
		return(throwtypeerror(b));
	}
	if(b->value.i<0) {
		DECREF(a);
		return(throwargumenterror(b));
	}
	result=a->value.i>>b->value.i;
	DECREF(a);
	DECREF(b);
	r=new_object(&integer_objtype);
	if(r)
		r->value.i=result;
	return(r);
}

obj *div2(obj *a,obj *b) {
	signed long result;
	obj *r;

	if(a->type != &integer_objtype) {
		DECREF(b);
		return(throwtypeerror(a));
	}
	if(b->type != &integer_objtype) {
		DECREF(a);
		return(throwtypeerror(b));
	}
	if(b->value.i==0) {
		DECREF(b);
		return(throw(INCREF(divideby0error),a));
	}
/* FIXME: Does this have undefined behavior on overflow? */
	result=a->value.i/b->value.i;
	DECREF(a);
	DECREF(b);
	r=new_object(&integer_objtype);
	if(r)
		r->value.i=result;
	return(r);
}

obj *mul(obj *a,obj *b) {
	signed long long result;
	int overflow=0;
	obj *r;

	if(a->type != &integer_objtype) {
		DECREF(b);
		return(throwtypeerror(a));
	}
	if(b->type != &integer_objtype) {
		DECREF(a);
		return(throwtypeerror(b));
	}
#ifdef RESTRICTED_MULTIPLICATION
	if(a->value.i<INT_MIN || INT_MAX<a->value.i) {
		DECREF(b);
		return(throwrangecheck(a));
	}
	if(b->value.i<INT_MIN || INT_MAX<b->value.i) {
		DECREF(a);
		return(throwrangecheck(b));
	}
	result=a->value.i*b->value.i;
#else
	result=a->value.i*(signed long long)b->value.i;
#endif
	if(result<LONG_MIN || LONG_MAX<result)
		overflow=1;
	DECREF(a);
	DECREF(b);
	r=new_object(&integer_objtype);
	if(r) {
		r->value.i=result;
		if(overflow)
			return(throw(INCREF(overflowerror),r));
	}
	return(r);
}

#define NUMCMP_LT 1
#define NUMCMP_EQ 2
#define NUMCMP_GT 4
obj *numcmp(obj *a,obj *b,int desired) {
	int cmp;

	if(a->type != &integer_objtype) {
		DECREF(b);
		return(throwtypeerror(a));
	}
	if(b->type != &integer_objtype) {
		DECREF(a);
		return(throwtypeerror(b));
	}
	cmp=NUMCMP_EQ;
	if(a->value.i < b->value.i)
		cmp=NUMCMP_LT;
	if(a->value.i > b->value.i)
		cmp=NUMCMP_GT;
	DECREF(a);
	DECREF(b);
	if(cmp&desired)
		return(INCREF(&trueobj));
	else
		return(INCREF(&nilobj));
}

obj *sub(obj *a,obj *b) {
	signed long result;
	int overflow=0;
	obj *r;

	if(a->type != &integer_objtype) {
		DECREF(b);
		return(throwtypeerror(a));
	}
	if(b->type != &integer_objtype) {
		DECREF(a);
		return(throwtypeerror(b));
	}
/* FIXME: This has undefined behavior on overflow: */
	result=a->value.i-b->value.i;
	if((a->value.i^b->value.i)<0 && (a->value.i^result)<0)
		overflow=1;
	DECREF(a);
	DECREF(b);
	r=new_object(&integer_objtype);
	if(r) {
		r->value.i=result;
		if(overflow)
			return(throw(INCREF(overflowerror),r));
	}
	return(r);
}

static obj *integer_tostring(obj *in) {
	obj *out;
	char buf[22];
	int i;

	i=snprintf(buf,sizeof(buf),"%ld",in->value.i);
	if(i==-1 || i>sizeof(buf))
		return(throwrangecheck(in));
	DECREF(in);
	out=new_string(buf,i);
	return(out);
}

static int integer_init() {
	if(!(divideby0error=new_word("divide-by-zero",14))
	|| !(overflowerror=new_word("overflow",8)))
		return(1);
/* Sanity check to make sure mul() will catch overflows.  If this fails for
 * you, you'll need to implement some method of overflow checking. */
#ifdef RESTRICTED_MULTIPLICATION
	if(sizeof(signed long) < 2*sizeof(signed int)) {
		fprintf(stderr,"long isn't twice as big as int on this machine\n");
		return(1);
	}
#else
	if(sizeof(signed long long) < 2*sizeof(signed long)) {
		fprintf(stderr,"longlong isn't twice as big as long on this machine\n");
		return(1);
	}
#endif
	return(0);
}

static void integer_exit() {
	DECREF(divideby0error); divideby0error=0;
	DECREF(overflowerror); overflowerror=0;
}

objtype integer_objtype={
	name: "integer",
	tostring: integer_tostring,
	init: integer_init,
	exit: integer_exit,
};

